import bisect
import hashlib
import os
from typing import NamedTuple, Optional
# local modules
from .protocol import V1Transfer, parse_v1_transfer, parse_v1_frame


class Chunk(NamedTuple):
    offset: int
    data: bytes

    def __lt__(self, other):
        return self.offset < other.offset

# not very robust against forged qr codes, but who cares :) I mean it is hashed, so I should be fine
class TransferReassembly:
    def __init__(self, data_hash: bytes):
        self.chunks: list[Chunk] = []
        self.hash = data_hash
        self.complete_data: Optional[bytes] = None

    def add_chunk(self, chunk: Chunk) -> bool:
        """Returns True, if the internal state was changed"""
        i = bisect.bisect_left(self.chunks, chunk)
        if i != len(self.chunks) and self.chunks[i].offset == chunk.offset:
            # a chunk with the same offset already exists
            old = self.chunks[i]
            if old.data == chunk.data:
                return False
            elif len(chunk.data) >= len(old.data):
                self.chunks[i] = chunk
        else:
            # no chunk with the same offset exists
            bisect.insort(self.chunks, chunk)

        self.complete_data = self.get_data_if_complete()
        return True

    def get_data_if_complete(self) -> Optional[bytes]:
        data = b""
        for chunk in self.chunks:
            # there should be no overlaps, so it should be irelevant which data I use
            start = chunk.offset
            end = start + len(chunk.data)
            if start <= len(data):
                if end > len(data):
                    data = data[:start] + chunk.data
            else:
                break

        hashed_data = hashlib.sha1(data).digest()
        return data if hashed_data == self.hash else None


class ReassemblyManager:
    def __init__(self, output_folder: str) -> None:
        self.transfers: dict[bytes,TransferReassembly] = {}
        self.output_folder = output_folder

    def add_data_chunk(self, data: bytes) -> None:
        frame = parse_v1_frame(data)
        transfer = self.transfers.get(frame.transfer_hash)
        if not transfer:
            transfer = TransferReassembly(frame.transfer_hash)
            self.transfers[frame.transfer_hash] = transfer

        chunk = Chunk(frame.transfer_offset, frame.data)
        if transfer.add_chunk(chunk):
            self.on_new_chunk(frame.transfer_hash, chunk)
            if transfer.complete_data:
                result = parse_v1_transfer(transfer.complete_data, frame.transfer_hash)
                self.on_transfer_completed(result)

    def on_new_chunk(self, transfer_hash: bytes, chunk: Chunk) -> None:
        end = chunk.offset + len(chunk.data) - 1
        print(f"{transfer_hash.hex()} | {chunk.offset}..{end}")

    def on_transfer_completed(self, transfer: V1Transfer):
        print(f"{transfer.transfer_hash.hex()} | Transfer finished: {transfer.name}")
        os.makedirs(self.output_folder, exist_ok=True)

        # The last check is just in case there is a system that has really strange path separators
        if "/" in transfer.name or "\\" in transfer.name or os.path.sep in transfer.name:
            # Path traversal detection
            print(f"[!] Security warning: Name '{transfer.name}' contains a path separator. This may be an attack! The transfer has been discarded")
        else:
            out_file = os.path.join(self.output_folder, transfer.name)
            with open(out_file, "wb") as f:
                f.write(transfer.contents)
            
            print(f"[*] Received file: {out_file} ({round(len(transfer.contents) / 1024.0, 2)} KiB)")



# Optional: Capture images via webcam (for air-gapped systems that you are pysically close to)
# Point your (external) webcam to the target's screen and show the QR codes on it
import os
import tempfile
import time
# local modules
from .assembler import ReassemblyManager
from .qr import parse_qr, MissingProgramException
# pip install opencv-python
import cv2
# # pip install Pillow
# from PIL import Image


# Based roughly on https://stackoverflow.com/questions/43776606/direct-way-to-get-camera-screenshot
def main_loop_with_webcam(assembler: ReassemblyManager, sleep_seconds: float, webcam_id: int):
    print(f"[*] Taking webcam images every {sleep_seconds} seconds. Open the web app at https://react-file2qr.vercel.app/ and start the QR slide show (or use qr2file). Exit receiving with Ctrl-C")
    print("[*] Hint: You will see output, when a new QR code containing a file segment is parsed")
    print("\n----------------- Hash ----------------- | --- Bytes received ---")

    cam = cv2.VideoCapture(webcam_id)
    _, qr_file = tempfile.mkstemp(suffix=".png")
    print(f"[*] The current webcam image is at {qr_file}. You can view this file with an image viewer to see what the camera is seeing and see if the camera id, positioning, etc are correct")
    try:
        while True:
            start = time.monotonic()
            # take and convert image
            _, img = cam.read()
            # img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            # im_pil = Image.fromarray(img)
            # We need to have it as a file anyways, so no directly converting it yould only save us one PNG read but complicate the code
            cv2.imwrite(qr_file, img)

            # parse image
            data_list = parse_qr(qr_file)
            for data in data_list:
                assembler.add_data_chunk(data)
            
            # delay next iteration
            diff = time.monotonic() - start
            remaining = sleep_seconds - diff
            if remaining > 0:
                time.sleep(remaining)
    except MissingProgramException as ex:
        print(f"\n\n[!] MissingProgramException: {ex}")
    except KeyboardInterrupt:
        print("<Ctrl-C>")
    finally:
        if os.path.exists(qr_file):
            os.remove(qr_file)


#!/usr/bin/env python3
import argparse
import os
import subprocess
import tempfile
import time
from typing import Any
# local modules
from .assembler import ReassemblyManager
from .qr import take_screenshot, parse_qr, MissingProgramException

DEFAULT_OUTPUT_FOLDER = "/tmp/qr-receiver/"


def notify(message: str, title: str) -> None:
    print("Notification:", message)
    subprocess.call(["notify-send", "--expire-time", "3000", title, message])


def parse_args() -> Any:
    ap = argparse.ArgumentParser()
    ap.add_argument("-s", "--sleep", type=int, help="the number of millis to sleep between screenshots", default=900)
    ap.add_argument("-o", "--output-dir", default=DEFAULT_OUTPUT_FOLDER, help=f"the folder to write the results to (default: {DEFAULT_OUTPUT_FOLDER})")
    ap.add_argument("-i", "--images", nargs="+", help="instead of taking regular screenshots, read all the given image files")
    ap.add_argument("-w", "--webcam", metavar="WEBCAM_ID", type=int, help="instead of taking screenshots, take pictures with a webcam. Usually 0 or 1 should be the ID of your first webcam")
    return ap.parse_args()

def main_loop_with_screenshots(assembler: ReassemblyManager, sleep_seconds: float):
    print(f"[*] Taking screenshots every {sleep_seconds} seconds. Open the web app at https://react-file2qr.vercel.app/ and start the QR slide show (or use qr2file). Exit receiving with Ctrl-C")
    print("[*] Hint: You will see output, when a new QR code containing a file segment is parsed")
    print("\n----------------- Hash ----------------- | --- Bytes received ---")

    _, qr_file = tempfile.mkstemp()
    try:
        while True:
            start = time.monotonic()
            take_screenshot(qr_file)
            data_list = parse_qr(qr_file)
            for data in data_list:
                assembler.add_data_chunk(data)
            diff = time.monotonic() - start
            remaining = sleep_seconds - diff
            if remaining > 0:
                time.sleep(remaining)
    except MissingProgramException as ex:
        print(f"\n\n[!] MissingProgramException: {ex}")
    except KeyboardInterrupt:
        print("<Ctrl-C>")
    finally:
        if os.path.exists(qr_file):
            os.remove(qr_file)


def main_loop_with_images(assembler: ReassemblyManager, image_file_list: list[str]):
    print(f"[*] Parsing {len(image_file_list)} image files")
    print("[*] Hint: You will see output, when a new QR code containing a file segment is parsed")
    print("\n----------------- Hash ----------------- | --- Bytes received ---")

    for image_file in image_file_list:
        if not os.path.exists(image_file):
            print(f"[-] File '{image_file}' does not exist")
        else:
            data_list = parse_qr(image_file)
            for data in data_list:
                assembler.add_data_chunk(data)

    print("[*] Done - all image files were processed")


def main():
    args = parse_args()
    if not os.path.isdir(args.output_dir):
        print(f"Output directory '{args.output_dir}' is not a directory or does not exist")
        exit(1)
    assembler = ReassemblyManager(args.output_dir)
    if args.images:
        image_file_list = args.images if type(args.images) == list else [args.images]
        main_loop_with_images(assembler, image_file_list)
    elif args.webcam is not None:
        # Conditional import. Only actually import opencv if we want/need to use it
        from .webcam import main_loop_with_webcam
        sleep_seconds = args.sleep / 1000
        main_loop_with_webcam(assembler, sleep_seconds, args.webcam)
    else:
        sleep_seconds = args.sleep / 1000
        main_loop_with_screenshots(assembler, sleep_seconds)


if __name__ == "__main__":
    main()



import os
import platform
import shutil
import subprocess
import tempfile
from typing import Optional
import warnings
# pip install Pillow
from PIL import Image
# pip install pyzbar
from pyzbar import pyzbar


def parse_qr(input_file: str) -> list[bytes]:
    if not os.path.exists(input_file):
        raise Exception(f"File '{input_file}' does not exist")

    # Hide the following warning I get when parsing images created with qrencode:
    #     /usr/lib/python3.11/site-packages/PIL/Image.py:970: UserWarning: Palette images with Transparency expressed in bytes should be converted to RGBA images
    #   warnings.warn(
    # @TODO: If someone knows how to fix/avoit it feel free to open an issue/PR
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        img = Image.open(input_file)
        qr_code_list = pyzbar.decode(img, symbols=[pyzbar.ZBarSymbol.QRCODE])
        qr_code_list = list(qr_code_list)

    if not qr_code_list:
        return []
    elif len(qr_code_list) == 1:
        return [read_binary_qr_code(input_file)]
    else:
        results = []
        for code in qr_code_list:
            code_contents = extract_and_read_qr_code(code, img)
            if code_contents:
                results.append(code_contents)
        return results


def extract_and_read_qr_code(code, img: Image) -> Optional[bytes]:
    _, tmp_file = tempfile.mkstemp()
    x, y, w, h = code.rect
    cropped = img.crop((x, y, x+w, y+h))
    # code.data is buggy, since it is null terminated. So i need to extract the QR codes and decode them individualy
    cropped.save(tmp_file)
    try:
        return read_binary_qr_code(tmp_file)
    except Exception as ex:
        print(ex)
        return None
    finally:
        os.remove(tmp_file)


def read_binary_qr_code(input_file: str) -> bytes:
    return subprocess.check_output(["zbarimg", "--quiet", "--raw", "-Sbinary", input_file])


def take_screenshot(output_file: str) -> None:
    if os.path.exists(output_file):
        os.remove(output_file)

    command = generate_screenshot_command(output_file, False)
    subprocess.call(command)

    if not os.path.exists(output_file):
        raise Exception(f"Command {command} did not produce output file {output_file}")

class MissingProgramException(Exception):
    pass


def generate_screenshot_command(output_file: str, user_selects_area: bool):
    # @SOURE/@SYNC https://gitlab.com/six-two/bin/-/blob/main/general/copy-qr-code
    os_type = platform.system()
    if os_type == "Linux":
        if os.getenv("XDG_SESSION_TYPE") == "wayland":
            if shutil.which("grim"):
                if user_selects_area:
                    try:
                        geometry = subprocess.getoutput("slurp")
                        return ["grim", "-g", geometry, output_file]
                    except Exception:
                        raise MissingProgramException("Failed to select region to screenshot using 'slurp'")
                else:
                    return ["grim", output_file]
            else:
                raise MissingProgramException("No linux wayland screenshotter found. Supported programs: grim")
        else:
            if shutil.which("scrot"):
                if user_selects_area:
                    return ["scrot", "-s", output_file]
                else:
                    return ["scrot", output_file]
            else:
                raise MissingProgramException("No linux X11 screenshotter found. Supported programs: scrot")
    elif os_type == "Darwin":
        if shutil.which("screencapture"):
            # This tool is installed on MacOS by default. You may need to allow 'Terminal' to record your screen (it should pop up a dialog)
            # Only tested in a single monitor setup, you may need to replace "-m" with something like "-D 2" to select the correct screen
            if user_selects_area:
                return ["screencapture", "-i", "-J", "selection", output_file]
            else:
                return ["screencapture", "-m", output_file]
        else:
            raise MissingProgramException("No MacOS screenshotter found. Supported programs: screencapture")
    else:
        raise MissingProgramException(f"Unknown system platform '{os_type}', expected 'Linux' or 'Darwin' (MacOS)")



#!/usr/bin/env python3
import argparse
import hashlib
import shutil
import subprocess
import time
from typing import Optional
import os


def int32_to_bytes(number: int) -> bytes:
    return number.to_bytes(4, 'big')

def transfer(data: bytes, name: str, chunk_size: int, delay_seconds: float, output_dir: Optional[str]) -> None:
    version = b"\x01"
    name_bytes = name.encode()
    name_bytes_len = int32_to_bytes(len(name_bytes))
    data_len = int32_to_bytes(len(data))
    serialized_file = version + name_bytes_len + name_bytes + data_len + data

    transfer_hash = hashlib.sha1(serialized_file).digest()
    print(f"[*] Transfering {transfer_hash.hex()} ({name}) - {len(data)} bytes")

    # @TODO delete old files (with same transfer id) when storing new images?

    for offset in range(0, len(serialized_file), chunk_size):
        offset_bytes = int32_to_bytes(offset)
        data_slice = serialized_file[offset:offset+chunk_size]
        frame_data = version + transfer_hash + offset_bytes + data_slice 

        if output_dir:
            # Store all QR codes in image files in the output directory
            if not os.path.isdir(output_dir):
                print(f"[!] '{output_dir}' is not a directory! Exiting...")
                exit(1)
            else:
                file_index = offset // chunk_size
                output_file_name = os.path.join(output_dir, f"{transfer_hash.hex()}_{file_index:03}.png")
                subprocess.run(["qrencode", "-m", "2", "-8", "-o", output_file_name], input=frame_data)
                # Show the progress. Each dot means a single file was generated
                print(".", end="", flush=True)
        else:
            # delete old qr code
            subprocess.call("clear")
            print(f"Sending {name} ({transfer_hash.hex()}): {offset+len(data_slice)}/{len(serialized_file)}")
            # show new qr code
            # -t ANSI256UTF8: smallest pixel representation
            # -m 2: smaller margin -> takes up less space
            # -8: needed for binary data to work
            subprocess.run(["qrencode", "-t", "ANSI256UTF8", "-m", "2", "-8"], input=frame_data)
            # wait before showing the next code
            time.sleep(delay_seconds)

    # Add a new-line after the dots being shown as the progress indicator
    if output_dir:
        print()


def main():
    if not shutil.which("qrencode"):
        print("[!] 'qrencode' program not found. It is needed to generate QR codes")
        exit(1)

    ap = argparse.ArgumentParser(description="Store the contents of a file in QR codes. This can be used to transfer a file between locked down citrix/RDP sessions or air-gapped systems. Use 'qr2file' to convert them back to the original file on the target system.")
    ap.add_argument("input_files", nargs="+", help="the file to transfer")
    ap.add_argument("-c", "--chunk-size", type=int, default=1000, help="how many file data to transmit per QR code")
    output_option_group = ap.add_mutually_exclusive_group()
    output_option_group.add_argument("-d", "--delay", type=float, default=1, help="how many seconds to wait between QR codes")
    output_option_group.add_argument("-o", "--output-dir", help="instead of showing QR codes, store them as image files in the given folder")
    args = ap.parse_args()

    # Check all files first so that it does not crash in the middle of transfering multiple files
    for input_file in args.input_files:
        if not os.path.exists(input_file):
            print(f"[!] '{input_file}' does not exist")
        if not os.path.isfile(input_file):
            print(f"[!] '{input_file}' is not a file")

    # Transfer the files
    try:
        for input_file in args.input_files:
            with open(input_file, "rb") as f:
                data = f.read()
            name = os.path.basename(input_file)
            transfer(data, name, args.chunk_size, args.delay, args.output_dir)
    except KeyboardInterrupt:
        print("\n[Ctrl-C] Stopped transfers")

if __name__ == "__main__":
    main()



from typing import NamedTuple

class ByteArrayReaderException(Exception):
    pass


class VersionException(Exception):
    def __init__(self, expected: int, actual: int) -> None:
        super().__init__(f"Version mismatch: expected {expected}, got {actual}")


class ByteArrayReader:
    def __init__(self, data: bytes) -> None:
        self.index = 0
        self.data = data
        if type(data) != bytes:
            raise Exception(f"Bad type: {type(data)}")

    def count_remaining(self) -> int:
        return len(self.data) - self.index

    def assert_remaining(self, min_count: int) -> None:
        actual = self.count_remaining()
        if actual < min_count:
            raise ByteArrayReaderException(f"Expected {min_count} byte(s), but only has {actual} remaining")

    def assert_finished(self) -> None:
        actual = self.count_remaining()
        if actual != 0:
            raise ByteArrayReaderException(f"Expected no more bytes, but only has {actual} remaining")

    def get_byte(self) -> int:
        self.assert_remaining(1)
        index = self.index
        self.index += 1
        return int(self.data[index])

    def get_bytes(self, length: int) -> bytes:
        self.assert_remaining(length)
        start = self.index
        self.index += length
        return self.data[start:self.index]

    def get_int32(self) -> int:
        data = self.get_bytes(4)
        value = 0
        value += data[0] << 24
        value += data[1] << 16
        value += data[2] << 8
        value += data[3]
        return value


class V1Frame(NamedTuple):
    version: int # 1 byte
    transfer_hash: bytes # 20 bytes
    transfer_offset: int # 4 bytes
    data: bytes # 1+ bytes


class V1Transfer(NamedTuple):
    version: int # 1 byte
    transfer_hash: bytes # 20 bytes (in each packet)
    name: str # 4 bytes + data
    contents: bytes # 4 bytes + data


def parse_v1_frame(payload: bytes) -> V1Frame:
    ba = ByteArrayReader(payload)
    version = ba.get_byte()
    if version != 1:
        raise VersionException(1, version)

    transfer_hash = ba.get_bytes(20)
    transfer_offset = ba.get_int32()
    data_len = ba.count_remaining()
    if data_len == 0:
        raise ByteArrayReaderException("No bytes for field data remaining")
    data = ba.get_bytes(data_len)
    ba.assert_finished()

    return V1Frame(version, transfer_hash, transfer_offset, data)


def parse_v1_transfer(payload: bytes, transfer_hash: bytes) -> V1Transfer:
    ba = ByteArrayReader(payload)
    version = ba.get_byte()
    if version != 1:
        raise VersionException(1, version)

    name_length = ba.get_int32()
    name = ba.get_bytes(name_length)
    contents_length = ba.get_int32()
    contents = ba.get_bytes(contents_length)
    ba.assert_finished()

    name_str = name.decode("utf-8")
    return V1Transfer(version, transfer_hash, name_str, contents)
